from __future__ import annotations

import dataclasses
import typing

import turbobt
import turbobt.neuron
import turbobt.subnet

from bt_ddos_shield.client import ShieldClient
from bt_ddos_shield.event_processor import PrintingMinerShieldEventProcessor
from bt_ddos_shield.turbobt.blockchain_manager import TurboBittensorBlockchainManager

if typing.TYPE_CHECKING:
    import bittensor_wallet

    from bt_ddos_shield.shield_metagraph import ShieldMetagraphOptions


class ShieldedBittensor(turbobt.Bittensor):
    ddos_shield: ShieldClient

    def __init__(
        self,
        *args,
        wallet: bittensor_wallet.Wallet,
        ddos_shield_netuid: int,
        ddos_shield_options: ShieldMetagraphOptions | None = None,
        **kwargs,
    ):
        super().__init__(
            *args,
            wallet=wallet,
            **kwargs,
        )

        event_processor = PrintingMinerShieldEventProcessor()

        self.ddos_shield = ShieldClient(
            ddos_shield_netuid,
            wallet,
            blockchain_manager=TurboBittensorBlockchainManager(
                self,
                netuid=ddos_shield_netuid,
                wallet=wallet,
                event_processor=event_processor,
            ),
            event_processor=event_processor,
            options=ddos_shield_options,
        )

    async def __aenter__(self):
        await super().__aenter__()
        await self.ddos_shield.__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.ddos_shield.__aexit__(*args, **kwargs)
        await super().__aexit__(*args, **kwargs)

    def subnet(self, netuid: int) -> turbobt.subnet.SubnetReference:
        if netuid == self.ddos_shield.netuid:
            return ShieldedSubnetReference(
                netuid,
                client=self,
            )

        return super().subnet(netuid)


class ShieldedSubnetReference(turbobt.subnet.SubnetReference):
    client: ShieldedBittensor = dataclasses.field(compare=False, repr=False)

    async def list_neurons(self, *args, **kwargs) -> list[turbobt.neuron.Neuron]:
        neurons = await super().list_neurons(*args, **kwargs)
        manifests = await self.client.ddos_shield.get_manifests([
            neuron.hotkey
            for neuron in neurons
        ])

        for neuron in neurons:
            manifest = manifests.get(neuron.hotkey)

            if not manifest:
                continue

            shield_address = self.client.ddos_shield.get_address(
                self.client.wallet.hotkey.ss58_address,
                manifest,
            )

            if shield_address is None:
                continue

            if self.client.ddos_shield.options.replace_ip_address_for_axon:
                neuron.axon_info.ip = shield_address[0]
            else:
                neuron.axon_info.shield_address = shield_address[0]

            neuron.axon_info.port = shield_address[1]

        return neurons



from __future__ import annotations

import typing

import turbobt

from bt_ddos_shield.blockchain_manager import AbstractBlockchainManager, BlockchainManagerException
from bt_ddos_shield.utils import decode_subtensor_certificate_info

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    import bittensor_wallet

    from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor
    from bt_ddos_shield.utils import Hotkey, PublicKey


class TurboBittensorBlockchainManager(AbstractBlockchainManager):
    def __init__(
        self,
        bittensor: turbobt.Bittensor,
        netuid: int,
        wallet: bittensor_wallet.Wallet,
        event_processor: AbstractMinerShieldEventProcessor,
    ):
        self.bittensor = bittensor
        self.wallet = wallet
        self.event_processor = event_processor

        self.subnet = turbobt.Bittensor.subnet(self.bittensor, netuid)

    def get_hotkey(self) -> Hotkey:
        return self.wallet.hotkey.ss58_address

    def get_own_public_key(self) -> PublicKey | None:
        raise NotImplementedError

    async def get_own_public_key_async(self) -> PublicKey | None:
        neuron = self.subnet.neuron(
            hotkey=self.get_hotkey(),
        )
        certificate = await neuron.get_certificate()

        if not certificate:
            return None

        certificate['public_key'] = [
            certificate['public_key'].encode(),
        ]

        decoded_certificate = decode_subtensor_certificate_info(certificate)

        if not decoded_certificate:
            return None

        return decoded_certificate.hex_data

    async def get_metadata(self, hotkeys: Iterable[Hotkey]) -> dict[Hotkey, bytes | None]:
        commitments = await self.subnet.commitments.fetch()
        commitments = {
            key: value
            for key, value in commitments.items()
            if key in hotkeys
        }

        for key in hotkeys:
            if key not in commitments:
                commitments[key] = None

        return commitments

    def put_metadata(self, data: bytes):
        raise NotImplementedError

    def upload_public_key(self, public_key: PublicKey):
        raise NotImplementedError

    async def upload_public_key_async(self, public_key: PublicKey) -> None:
        try:
            neuron = await self.subnet.get_neuron(self.get_hotkey())

            ip = '1.1.1.1'
            port = 1
            protocol = 0

            if neuron and neuron.axon_info:
                ip = str(neuron.axon_info.ip or ip)
                port = neuron.axon_info.port or port
                protocol = neuron.axon_info.protocol or protocol

            await self.subnet.neurons.serve(
                ip,
                port,
                certificate=bytes.fromhex(public_key),
                wallet=self.wallet,
            )
        except Exception as e:
            self.event_processor.event(
                'Failed to upload public key for netuid={netuid}, wallet={wallet}',
                exception=e,
                netuid=self.subnet.netuid,
                wallet=self.wallet,
            )
            raise BlockchainManagerException(f'Failed to upload public key: {e}') from e



from __future__ import annotations

import asyncio
import copy
from typing import TYPE_CHECKING

import pytest

from bt_ddos_shield.address_manager import AwsAddressManager
from bt_ddos_shield.miner_shield import MinerShield, MinerShieldFactory
from bt_ddos_shield.shield_metagraph import ShieldMetagraph
from bt_ddos_shield.state_manager import SQLAlchemyMinerShieldStateManager
from bt_ddos_shield.validators_manager import BittensorValidatorsManager

if TYPE_CHECKING:
    from bittensor.core.chain_data import AxonInfo

    from tests.conftest import ShieldTestSettings


@pytest.fixture
def miner_shield(shield_settings: ShieldTestSettings):
    # We need to add any validator to set, otherwise manifest addresses will be created for all validators in
    # network including tested validator, what we don't want
    validators = {'unknown_hotkey'}

    shield = MinerShieldFactory.create_miner_shield(
        shield_settings,
        validators,
    )

    assert isinstance(shield.address_manager, AwsAddressManager)
    assert isinstance(shield.state_manager, SQLAlchemyMinerShieldStateManager)
    assert isinstance(shield.validators_manager, BittensorValidatorsManager)

    shield.state_manager.clear_tables()
    shield.enable()
    assert shield.run

    # Wait for full shield initialization - should create empty manifest
    shield.task_queue.join()

    yield shield

    shield.disable()
    assert not shield.run
    shield.address_manager.clean_all()


class TestValidator:
    """
    Test suite for the Validator class.
    """

    def test_full_flow(self, miner_shield: MinerShield, shield_settings: ShieldTestSettings):
        """
        Test if validator is working using real managers and real shield.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        miner_hotkey: str = shield_settings.wallet.instance.hotkey.ss58_address

        metagraph: ShieldMetagraph = ShieldMetagraph(
            wallet=shield_settings.validator_wallet.instance,
            subtensor=shield_settings.subtensor.create_client(),
            netuid=shield_settings.netuid,
        )
        miner_axon: AxonInfo = next(axon for axon in metagraph.axons if axon.hotkey == miner_hotkey)

        miner_shield.disable()
        miner_shield.validators_manager.validators = frozenset()
        miner_shield.enable()
        miner_shield.task_queue.join()  # Wait for full shield initialization - should add validator to manifest

        metagraph.sync()
        shielded_miner_axon: AxonInfo = next(axon for axon in metagraph.axons if axon.hotkey == miner_hotkey)
        assert shielded_miner_axon.ip != miner_axon.ip
        assert shielded_miner_axon.port == miner_shield.address_manager.ELB_LISTENING_PORT

    def test_full_flow_in_async_context(self, shield_settings: ShieldTestSettings):
        async def async_wrapper():
            self.test_full_flow(shield_settings)

        asyncio.run(async_wrapper())

    def test_copy(self, shield_settings: ShieldTestSettings):
        metagraph: ShieldMetagraph = ShieldMetagraph(
            wallet=shield_settings.validator_wallet.instance,
            subtensor=shield_settings.subtensor.create_client(),
            netuid=shield_settings.netuid,
        )

        metagraph_copy: ShieldMetagraph = copy.deepcopy(metagraph)
        assert metagraph_copy.axons == metagraph.axons
        assert metagraph.subtensor is not None
        assert metagraph_copy.subtensor is None, 'Metagraph class ignores subtensor field during deepcopy'

        metagraph_copy = copy.copy(metagraph)
        assert metagraph_copy.axons == metagraph.axons
        assert metagraph.subtensor is not None
        assert metagraph_copy.subtensor is None, 'Metagraph class ignores subtensor field during copy'


class TestTurboBittensor:
    """
    Test suite for turbobt's ShieldedBittensor.
    """

    @pytest.fixture(autouse=True, scope='class')
    def importorskip(self):
        pytest.importorskip('turbobt')

    @pytest.mark.asyncio
    async def test_full_flow(self, miner_shield, shield_settings: ShieldTestSettings):
        """
        Test if validator is working using real managers and real shield.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        from bt_ddos_shield.turbobt import ShieldedBittensor

        miner_hotkey: str = shield_settings.wallet.instance.hotkey.ss58_address

        async with ShieldedBittensor(
            shield_settings.subtensor.network,
            ddos_shield_netuid=shield_settings.netuid,
            wallet=shield_settings.validator_wallet.instance,
        ) as bittensor:
            subnet = bittensor.subnet(shield_settings.netuid)
            neurons = await subnet.list_neurons()

            miner = next(neuron for neuron in neurons if neuron.hotkey == miner_hotkey)

            miner_shield.disable()
            miner_shield.validators_manager.validators = frozenset()
            miner_shield.enable()

            # Wait for full shield initialization - should add validator to manifest
            miner_shield.task_queue.join()

            neurons = await subnet.list_neurons()

            shielded_miner = next(neuron for neuron in neurons if neuron.hotkey == miner_hotkey)

            assert shielded_miner.axon_info.ip != miner.axon_info.ip
            assert shielded_miner.axon_info.port == miner_shield.address_manager.ELB_LISTENING_PORT



from __future__ import annotations

import asyncio
import os
import typing

from bt_ddos_shield.blockchain_manager import BlockchainManagerException
from bt_ddos_shield.encryption_manager import ECIESEncryptionManager, EncryptionCertificate
from bt_ddos_shield.manifest_manager import (
    JsonManifestSerializer,
    Manifest,
    ManifestDeserializationException,
    ReadOnlyManifestManager,
)
from bt_ddos_shield.shield_metagraph import ShieldMetagraphOptions

if typing.TYPE_CHECKING:
    import bittensor_wallet

    from bt_ddos_shield.blockchain_manager import AbstractBlockchainManager
    from bt_ddos_shield.encryption_manager import AbstractEncryptionManager
    from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor
    from bt_ddos_shield.utils import Hotkey


class ShieldClient:
    blockchain_manager: AbstractBlockchainManager
    certificate: EncryptionCertificate
    encryption_manager: AbstractEncryptionManager
    event_processor: AbstractMinerShieldEventProcessor
    manifest_manager: ReadOnlyManifestManager
    options: ShieldMetagraphOptions
    wallet: bittensor_wallet.Wallet

    def __init__(
        self,
        netuid: int,
        wallet: bittensor_wallet.Wallet,
        event_processor: AbstractMinerShieldEventProcessor,
        blockchain_manager: AbstractBlockchainManager,
        encryption_manager: AbstractEncryptionManager | None = None,
        manifest_manager: ReadOnlyManifestManager | None = None,
        options: ShieldMetagraphOptions | None = None,
    ):
        self.netuid = netuid
        self.wallet = wallet
        self.options = options or ShieldMetagraphOptions()
        self.event_processor = event_processor
        self.blockchain_manager = blockchain_manager
        self.encryption_manager = encryption_manager or self.create_default_encryption_manager()
        self.manifest_manager = manifest_manager or self.create_default_manifest_manager(
            self.event_processor,
            self.encryption_manager,
        )

    async def __aenter__(self):
        await self._init_certificate()

    async def __aexit__(self, *args, **kwargs):
        pass

    @classmethod
    def create_default_encryption_manager(cls):
        return ECIESEncryptionManager()

    @classmethod
    def create_default_manifest_manager(
        cls,
        event_processor: AbstractMinerShieldEventProcessor,
        encryption_manager: AbstractEncryptionManager,
    ) -> ReadOnlyManifestManager:
        return ReadOnlyManifestManager(JsonManifestSerializer(), encryption_manager, event_processor)

    async def _init_certificate(self) -> None:
        certificate_path: str = self.options.certificate_path or os.getenv(
            'VALIDATOR_SHIELD_CERTIFICATE_PATH', './validator_cert.pem'
        )

        try:
            certificate = self.encryption_manager.load_certificate(certificate_path)
            self.certificate = self.encryption_manager.serialize_certificate(certificate)
            public_key = await self.blockchain_manager.get_own_public_key_async()

            if self.certificate.public_key == public_key:
                return
        except FileNotFoundError:
            certificate = self.encryption_manager.generate_certificate()
            self.encryption_manager.save_certificate(certificate, certificate_path)
            self.certificate = self.encryption_manager.serialize_certificate(certificate)

        if self.options.disable_uploading_certificate:
            return

        try:
            await self.blockchain_manager.upload_public_key_async(self.certificate.public_key)
        except BlockchainManagerException:
            # Retry once
            await asyncio.sleep(3)
            await self.blockchain_manager.upload_public_key_async(self.certificate.public_key)

    async def get_manifests(
        self,
        hotkeys: typing.Iterable[Hotkey],
    ) -> dict[Hotkey, Manifest | None]:
        manifests_urls = await self.blockchain_manager.get_manifest_urls(hotkeys)
        manifests = await self.manifest_manager.get_manifests(manifests_urls)

        return manifests

    def get_address(
        self,
        hotkey: Hotkey,
        manifest: Manifest,
    ) -> tuple[str, int] | None:
        try:
            return self.manifest_manager.get_address_for_validator(
                manifest,
                hotkey,
                self.certificate.private_key,
            )
        except ManifestDeserializationException as e:
            self.event_processor.event(
                'Error while getting shield address for miner {hotkey}',
                exception=e,
                hotkey=hotkey,
            )
            return None
